#include <sys/types.h>
#include <sys/mman.h>

#include <machine/vmm.h>

#include <stdio.h>
#include <libgen.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <vmmapi.h>

static const char *progname;

struct devinfo {
	size_t	size;
	int	segid;
	int	prot;
	uint64_t gpa;
	const char *name;
};

static const struct devinfo defaults = {
	.size	= 0x200000,	/* 2MB */
	.segid	= 1,
	.prot	= PROT_READ | PROT_WRITE,
	.gpa	= 0xE0000000,	/* 3.5GB */
	.name	= "devtest"
};

static void
usage(int rc)
{

	fprintf(stderr, "Usage: %s [-chmN][-s <size>][-i <segid>][-n <devname>]"
	    "[-p <prot>][-g <gpa>] vmname\n"
	    "        -c: create the devmem segment\n"
	    "        -h: print this help message\n"
	    "        -m: mmap devmem in guest address space\n"
	    "        -s: size of the devmem segment (default is %#lx)\n"
	    "        -i: devmem segment id (default is %d)\n"
	    "        -n: devmem segment name (default is %s)\n"
	    "        -N: set devmem segment name to NULL\n"
	    "        -g: address where devmem is mapped (default is %#lx)\n"
	    "        -p: mapping protection bits (default is %c%c%c)\n",
	    progname, defaults.size, defaults.segid, defaults.name,
	    defaults.gpa, defaults.prot & PROT_READ ? 'r' : '-',
	    defaults.prot & PROT_WRITE ? 'w' : '-',
	    defaults.prot & PROT_EXEC ? 'x' : '-');

	exit(rc);
}

int
main(int argc, char *argv[])
{
	struct devinfo di;
	struct vmctx *ctx;
	const char *vmname;
	void *ptr;
	int c, create, error, map;

	progname = basename(argv[0]);
	bcopy(&defaults, &di, sizeof(struct devinfo));
	create = map = 0;

	while ((c = getopt(argc, argv, "chmNs:i:n:g:p:")) != -1) {
		switch (c) {
		case 'c':
			create = 1;
			break;
		case 'm':
			map = 1;
			break;
		case 's':
			di.size = strtoul(optarg, NULL, 0);
			break;
		case 'i':
			di.segid = (int)strtol(optarg, NULL, 0);
			break;
		case 'n':
			di.name = strdup(optarg);
			break;
		case 'N':
			di.name = NULL;
			break;
		case 'g':
			di.gpa = strtoul(optarg, NULL, 0);
			break;
		case 'p':
			di.prot = (int)strtol(optarg, NULL, 0);
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage(1);

	errno = 0;

	vmname = argv[0];
	error = vm_create(vmname);
	if (error && errno != EEXIST) {
		perror("vm_create");
		exit(1);
	}

	ctx = vm_open(vmname);
	if (ctx == NULL) {
		perror("vm_open");
		exit(1);
	}

	if (create) {
		ptr = vm_create_devmem(ctx, di.segid, di.name, di.size);
		if (ptr == MAP_FAILED) {
			perror("vm_create_devmem");
			exit(1);
		}
	}

	if (map) {
		error = vm_mmap_memseg(ctx, di.gpa, di.segid, 0, di.size,
		    di.prot);
		if (error) {
			perror("vm_mmap_devmem");
			exit(1);
		}
	}

	exit(0);
}
